import {createDriverAPI, findNode} from '@servicenow/ui-testing-helpers';
import {createNowIconDriver} from '@servicenow/now-icon/drivers';
import {createNowAvatarDriver} from '@servicenow/now-avatar/drivers';

export const createNowDropdownListDriver = createDriverAPI(
	'now-dropdown-list',
	async ({find, element, driver, node}) => {
		const componentId = await driver.getAttribute('component-id');
		const __private = {
			async getAllItems() {
				return await element('>>> .now-dropdown-list-item', node, true);
			},

			async getItem(matcher) {
				const nodes = await __private.getAllItems();
				const exactMatch = await findNode({
					element,
					matcher,
					nodes
				});

				if (exactMatch) {
					return exactMatch;
				}

				// Panel items are prefixed with the component-id, which might be
				// auto-generated. See if we can find a panel item using the ID
				// originally passed in the props (without any prefix)
				return findNode({
					element,
					matcher,
					nodes,
					getId: async (el) => {
						const itemId = await el.getAttribute('id');
						return itemId.slice(0, -(componentId.length + 1));
					}
				});
			},

			async getHighlights(sublabelOnly) {
				const items = await __private.getAllItems();
				return element(
					`${sublabelOnly ? '.now-dropdown-list-sublabel' : ''} .has-highlight`,
					await items.getNodes(),
					true
				);
			},

			async getIcon(matcher) {
				const item = await __private.getItem(matcher);
				if (item) {
					const icon = await item.getNodes('now-icon,now-icon-presence');
					return createNowIconDriver(icon, element);
				}

				return element();
			},

			async getAvatar(matcher) {
				const item = await __private.getItem(matcher);
				if (item) {
					const icon = await item.getNodes('now-avatar');
					return createNowAvatarDriver(icon, element);
				}

				return element();
			},

			// TODO: When using old fit behavior, Firefox is 1px off (the itemBottom
			// is 1px below the scrollBottom). Didn't need this before, or with new
			// fit... so we should get rid of this fuzz param at some point.
			async itemIsInScrollView(matcher, fuzz = 1) {
				const itemEl = await __private
					.getItem(matcher)
					.then((item) => item.getNodes());
				const listbox = itemEl.parentNode;
				const scrollTop = listbox.scrollTop;
				const scrollBottom = scrollTop + listbox.clientHeight;
				const itemTop = itemEl.offsetTop;
				const itemBottom = itemTop + itemEl.clientHeight;
				return scrollTop - fuzz <= itemTop && scrollBottom + fuzz >= itemBottom;
			},

			async isFocused(matcher) {
				const item = await __private.getItem(matcher);
				if (!item) {
					return false;
				}

				return item.isFocused();
			},

			async countHeaders() {
				return await driver.count('>>> .now-dropdown-list-header');
			},

			async countDividers() {
				return await driver.count('>>> .now-dropdown-list-divider');
			},

			async getListContainer() {
				return await driver.getNodes('>>> .now-dropdown-list-item-container');
			},

			async getSearchWrapper() {
				return await driver.getNodes('>>> .now-dropdown-list-search');
			},

			async getSearchField() {
				return await driver.getNodes('>>> .now-dropdown-list-search-field');
			}
		};

		return {
			__private,

			/**
			 * Returns `true` if the panel is displayed
			 * @returns {Promise<boolean>}
			 */
			async isDisplayed() {
				return (await find(`.now-dropdown-list`)).isDisplayed();
			},

			/**
			 * Returns the number of items in the dropdown panel.
			 * @returns {Promise<number>}
			 */
			async count() {
				const nodes = await __private.getAllItems();
				return nodes.count();
			},

			/**
			 * Selects the first matching item in the panel.
			 * @param {Matcher} matcher
			 * @returns {Promise<void>}
			 */
			async selectItem(matcher) {
				const item = await __private.getItem(matcher);
				if (item) {
					return item.click();
				}
			},

			/**
			 * Returns `true` if the matching item in the panel is selected.
			 * @param {Matcher} matcher
			 * @returns {Promise<boolean>}
			 */
			async isSelected(matcher) {
				const item = await __private.getItem(matcher);
				if (!item) {
					return false;
				}
				const selected = await item.getAttribute('aria-checked');
				const icon = await element(
					'now-icon[icon=check-fill]',
					await item.getNodes()
				);
				return selected === 'true' && (await icon.isExisting());
			},

			/**
			 * Returns the number of selected items in the dropdown panel.
			 * @returns {Promise<number>}
			 */
			async countSelected() {
				const selected = await element(
					'>>> now-icon[icon=check-fill]',
					await driver.getNodes(),
					true
				);
				return selected.count();
			}
		};
	}
);
