import {createEnhancedElement} from '@servicenow/library-enhanced-element';
import focusBehavior from '@servicenow/behavior-focus';
import {makeRedispatchEffect} from '@servicenow/library-enhanced-utils';
import '@servicenow/now-popover';
import './now-dropdown-list';
import {itemsSchema, itemSchema} from './schemas/dropdown-schema';
import {constrainSchema} from '@servicenow/behavior-fit';
import {pick} from 'lodash';
import {setActiveItem, getDropdownListConstrain} from './utils/helpers.js';
import {configAriaSchema} from '@servicenow/library-schemas';

const assignedElements = (slot) =>
	[...slot.assignedNodes({flatten: true})].filter(
		(node) => node.nodeType === Element.ELEMENT_NODE
	);

const view = (state, {updateState}) => {
	const {targetRef, activeItem} = state;
	const {
		items,
		selectedItems,
		select,
		opened,
		panelFitProps,
		search,
		configAria
	} = state.properties;

	const constrain = {
		minHeight: 32,
		minWidth: 'target',
		maxWidth: 400,
		maxHeight: 300,
		...panelFitProps.constrain
	};

	return (
		<now-popover
			hide-tail
			constrain={pick(constrain, ['width', 'minWidth', 'maxWidth'])}
			positions={panelFitProps.position}
			offset={panelFitProps.offset}
			manage-opened
			opened={opened}
			position-target={targetRef}
			interaction-type="listbox"
			disable-scroll>
			<slot
				className="now-dropdown-custom-target-slot"
				name="trigger"
				slot="trigger"
				on-slotchange={(e) => {
					if (e.target !== e.currentTarget) {
						return; // ignore bubbling slotchange
					}
					const elements = assignedElements(e.target);
					const targetRef = elements[0] || null;
					updateState({targetRef});
				}}
			/>
			<now-dropdown-list
				slot="content"
				style={getDropdownListConstrain(constrain)}
				items={items}
				selected-items={selectedItems}
				select={select}
				search={search}
				active-item={activeItem}
				manage-active-item
				manage-selected-items
				config-aria={configAria}
			/>
		</now-popover>
	);
};

/**
 * A dropdown with a custom target is very similar to a standard dropdown.
 * However, the trigger element (the element which toggles the panel
 * visibility) is provided via a slot, allowing more control of the overall
 * appearance.
 *
 * ```jsx
 * const dropdownItems = [{id: 'item1', label: 'Option 1'}, {id: 'item2', label: 'Option 2'}];
 *
 * <now-dropdown-custom-target items={dropdownItems}>
 *   <now-avatar user-name="Fred Luddy" slot="target" />
 * </now-dropdown-custom-target>
 * ```
 *
 * @seismicElement now-dropdown-custom-target
 * @summary A dropdown that allows user to select one or more values and
 * control the element that is responsible for opening and closing the panel.
 **/
createEnhancedElement('now-dropdown-custom-target', {
	initialState: {
		targetRef: null
	},
	properties: {
		/**
		 * By default, the component will listen for click & key events on the
		 * element in the `trigger` slot to cause the dropdown panel to open,
		 * and will restore focus to the `trigger` element when the panel is
		 * closed (on selection or press of the `Escape` key). Set this flag
		 * to override the default behavior and respond to user interactions
		 * manually instead.
		 * @type {boolean}
		 */
		disableListeners: {default: false, schema: {type: 'boolean'}},
		/**
		 * An array of item objects describing each item in the dropdown.
		 * Each object must have a unique id and label.
		 * Items with `disabled` will be visible but not selectable by the user.
		 * @type {(Array.<DropdownItem> | Array.<DropdownSection>)}
		 */
		items: {
			required: true,
			schema: itemsSchema
		},
		/**
		 * An array of ids representing each selected item. If multi-select mode is
		 * not enabled, each time the user selects an item, the previously selected
		 * id will be cleared out and the new id will be added. If multi-select mode is
		 * enabled, ids will be added to or removed from the list.
		 * Use `manage-selected-items` to override the default behavior and handle the
		 * `NOW_DROPDOWN#SELECTED_ITEMS_SET` action manually.
		 * @type {Array.<(string|number)>}
		 */
		selectedItems: {
			default: [],
			manageable: true,
			schema: {type: 'array', items: {type: ['string', 'number']}}
		},
		/**
		 * Determines what happens when the user clicks on a dropdown item.
		 * Choose from the following:
		 * - "single" (default) - the panel is closed, the trigger label is updated,
		 *   and the item is marked as selected
		 * - "multi" - the panel stays open, the trigger label is updated, and the
		 *   item's selected state is toggled on/off
		 * - "none" - the panel is closed, the trigger label is not updated, the
		 *   item is not marked as selected
		 * For select "single"/"multi" handle the `NOW_DROPDOWN#SELECTED_ITEMS_SET`
		 * action. For select "none" handle the `NOW_DROPDOWN#ITEM_CLICKED` action.
		 * @type {("single"|"multi"|"none")}
		 */
		select: {
			default: 'single',
			schema: {type: 'string', enum: ['single', 'multi', 'none']}
		},
		/**
		 * Automatically updated when the user opens or closes the dropdown.
		 * Use `manage-opened` to override the default behavior and handle the
		 * `NOW_DROPDOWN#OPENED_SET` action manually.
		 * @type {boolean}
		 */
		opened: {default: false, manageable: true, schema: {type: 'boolean'}},
		/**
		 * Properties to configure the dropdown panel.
		 *
		 * Valid keys are:
		 *	- position,
		 *	- offset
		 *	- container
		 *	- constrain
		 *
		 * Each of these keys are used to control how the panel is positioned and
		 * scaled relative to the dropdown target.
		 *
		 * @type {{ position: Array<string>, offset: number, container: HTMLElement, constrain: object }}
		 */
		panelFitProps: {
			default: {},
			schema: {
				type: 'object',
				properties: {
					position: {type: 'array', items: {type: 'string'}},
					offset: {
						oneOf: [{type: 'number'}, {type: 'array', items: {type: 'number'}}]
					},
					container: {type: 'object'},
					constrain: constrainSchema
				}
			}
		},
		/**
		 * Controls how to filter the items as the user types into the integrated search input.
		 * - "none": (default) dropdown list does not display an internal search field.
		 * - "managed": internal input field is rendered, but no automatic filtering will happen.
		 * Use this if you want to manually filter or search for items as the user types outside of the list.
		 * - "initial": internal input field is rendered, and items whose label
		 * or sublabel begins with the input value will be displayed. This is case-insensitive.
		 * - "contains": internal input field is rendered, and items whose label
		 * or sublabel contains the input value will be displayed. This is case-insensitive.
		 * @type {("none"|"managed"|"initial"|"contains")}
		 * @uib.label Search Behavior
		 * @uib.description Controls how to filter the items as the user types into the integrated search input
		 */
		search: {
			default: 'none',
			schema: {
				type: 'string',
				enum: ['none', 'managed', 'initial', 'contains']
			}
		},
		/**
		 * An object whose keys reference a specific HTML element within the
		 * now-dropdown-custom-target. The value of each key is an object that contains an ARIA
		 * property (or properties) and corresponding value/s. These ARIA
		 * properties will be set on the specified element.
		 * The 'panel' key corresponds to the inner html `<div>` that acts as the
		 * 'listbox'. See https://www.w3.org/TR/wai-aria-1.1/#listbox for properties
		 * and accepted values.
		 * @type {{ 'aria-*': string }}
		 * @uib.label ARIA Properties
		 * @uib.description Sets ARIA properties on the inner div with role 'listbox'.
		 */
		configAria: {
			schema: {
				anyOf: [
					{
						type: 'object',
						properties: {
							panel: configAriaSchema
						}
					},
					configAriaSchema
				]
			}
		}
	},
	dispatches: {
		/**
		 * Dispatched when the dropdown menu is opened or closed. Payload will contain
		 * a boolean of `true` when the dropdown is opened or `false` when the dropdown is closed.
		 * Set the `manage-opened` property to override the default behavior and
		 * handle this action manually.
		 * @type {{value: boolean}}
		 */
		'NOW_DROPDOWN_CUSTOM_TARGET#OPENED_SET': {
			schema: {
				type: 'object',
				properties: {
					value: {type: 'boolean'}
				}
			}
		},
		/**
		 * Dispatched when the user changes the selection of items in the dropdown panel.
		 * Payload key `value` will contain an array containing either a single ID (for `select="single"`) or
		 * multiple IDs (for `select="multi"`) of the selected item(s).
		 * Payload key `items` contains an array of objects of the item or items selected from the dropdown list.
		 * Set the `manage-selected-items` property to override the default behavior and
		 * handle this action manually.
		 * @type {{value: Array.<(string|number)>, items: Array.<Object>}}
		 */
		'NOW_DROPDOWN_CUSTOM_TARGET#SELECTED_ITEMS_SET': {
			schema: {
				type: 'object',
				properties: {
					value: {
						type: 'array',
						items: {
							type: ['string', 'number']
						}
					}
				}
			}
		},
		/**
		 * Dispatched when the user clicks on a dropdown item. Payload will contain
		 * a reference to the clicked item. Handle this action for dropdown menus (where `select="none"`).
		 * @type {{item: Object}}
		 */
		'NOW_DROPDOWN_CUSTOM_TARGET#ITEM_CLICKED': {
			schema: itemSchema
		}
	},
	slots: {
		/**
		 * Content placed in this named slot will act as the trigger element for
		 * the `now-dropdown-custom-target` component. The dropdown panel will
		 * be positioned relative to the first child element of the slot. Clicks
		 * and keyboard interaction with the slot and its content will be
		 * handled automatically unless the `disableListeners` property is set.
		 */
		trigger: {}
	},
	eventHandlers: [
		{
			events: ['keydown'],
			effect: ({action, properties, state}) => {
				if (properties.disableListeners) {
					return;
				}
				const {event} = action.payload;
				const openKeys = ['ArrowUp', 'ArrowDown', 'Enter', ' '];

				if (openKeys.includes(event.key)) {
					const {targetRef} = state;
					action.preventDefault();
					targetRef.click();
				}
			}
		}
	],
	actionHandlers: {
		'NOW_POPOVER#OPENED_SET': ({action, dispatch, properties, updateState}) => {
			const {value} = action.payload;
			const {disabled, items, selectedItems, disableListeners} = properties;
			if (disabled || disableListeners) return;
			if (value) {
				setActiveItem(items, selectedItems, updateState);
			}
			dispatch(() => {
				return {
					type: 'NOW_DROPDOWN_CUSTOM_TARGET#OPENED_SET',
					payload: action.payload
				};
			});
		},
		'NOW_POPOVER#CONTENT_VISIBLE': ({host}) => {
			const dropdownList = host.shadowRoot.querySelector('now-dropdown-list');
			dropdownList.focus();
		},
		'NOW_DROPDOWN_LIST#SELECTED_ITEMS_SET': makeRedispatchEffect(
			'NOW_DROPDOWN_CUSTOM_TARGET#SELECTED_ITEMS_SET'
		),
		'NOW_DROPDOWN_LIST#ITEM_CLICKED': ({dispatch, action}) => {
			dispatch(({properties}) => {
				const {select} = properties;
				return {
					type: 'NOW_DROPDOWN_CUSTOM_TARGET#OPENED_SET',
					payload: {value: false},
					shouldDispatch: select !== 'multi'
				};
			});
			dispatch(() => ({
				type: 'NOW_DROPDOWN_CUSTOM_TARGET#ITEM_CLICKED',
				payload: {item: action.payload.item}
			}));
		},
		'NOW_DROPDOWN_LIST#ACTIVE_ITEM_SET': ({action, updateState}) => {
			const activeItem = action.payload.value;
			updateState({activeItem});
		}
	},
	behaviors: [{behavior: focusBehavior}],
	view
});
