import {createEnhancedElement} from '@servicenow/library-enhanced-element';
import {Fragment} from '@servicenow/ui-renderer-snabbdom';
import {actionTypes} from '@servicenow/ui-core';
import styles from './_now-dropdown-panel.scss';
import {
	addOutsideScrollListener,
	removeOutsideScrollListener
} from './panel-outside-scroll.js';
import {getBestFitInfo} from '@servicenow/behavior-fit';
import {t} from '@servicenow/library-translate';
import isEqual from 'lodash/isEqual';
import '@servicenow/now-icon';
import {itemsSchema} from './schemas/dropdown-schema';
import {constrainSchema} from '@servicenow/behavior-fit';
import {filterAriaAttributes} from '@servicenow/library-enhanced-utils';
import {configAriaSchema} from '@servicenow/library-schemas';

const {
	COMPONENT_DOM_READY,
	COMPONENT_RENDERED,
	COMPONENT_PROPERTY_CHANGED
} = actionTypes;

/** copy from ./utils/elements.js to isolate panel code*/
function createIdSelector(prefix) {
	return function(host) {
		const componentId = host.getAttribute('component-id');
		return host.shadowRoot.getElementById(`${prefix}-${componentId}`);
	};
}

const getPanelEl = createIdSelector('panel');

const getContentEl = createIdSelector('panel');

const getListboxEl = (host) =>
	host.shadowRoot.querySelector('.now-dropdown-panel-list');

const getItemEls = (host) => {
	return [...host.shadowRoot.querySelectorAll('.now-dropdown-panel-item')];
};

const canRenderDivider = (index, array) => {
	const isLastItem = index === array.length - 1;
	if (isLastItem) {
		return false;
	}

	const nextItem = array[index + 1];
	const isNextSectionWithLabel =
		nextItem.label && Array.isArray(nextItem.children);

	return !isNextSectionWithLabel;
};

const getSelectedItemEls = (host) => {
	return [
		...host.shadowRoot.querySelectorAll(
			'.now-dropdown-panel-item[aria-selected="true"]'
		)
	];
};

const getFocusedItemEl = (host) => {
	return host.shadowRoot.activeElement;
};

const getFocusedItemIndex = (host) => {
	return getItemEls(host).findIndex(
		(element) => element.id === getFocusedItemEl(host).id
	);
};

const getShadowRootHost = (el) => {
	if (el.parentElement) {
		return getShadowRootHost(el.parentElement);
	}
	if (el.parentNode && el.parentNode instanceof DocumentFragment) {
		return el.parentNode.host;
	}
	return null;
};

const isPanelList = (el) =>
	el.getAttribute('role') === 'listbox' || el.getAttribute('role') === 'menu';
// end of utils

/** copy from ./utils/focus.js to isolate panel code*/
const unfocusItem = (host) => {
	const listbox = getListboxEl(host);
	listbox.focus();
};

const focusOnOpen = async (host) => {
	const listbox = getListboxEl(host);
	listbox.focus();
	const firstFocusItem = getSelectedItemEls(host)[0] || getItemEls(host)[0];
	if (firstFocusItem) {
		firstFocusItem.focus();
	}
};

const focusPreviousItem = (host) => {
	const lastFocusItem = getFocusedItemEl(host);
	if (!lastFocusItem || isPanelList(lastFocusItem)) {
		focusFirstItem(host);
		return;
	}
	const lastFocusIndex = getFocusedItemIndex(host);
	const nextFocusItem = getItemEls(host)[lastFocusIndex - 1];
	if (nextFocusItem) {
		nextFocusItem.focus();
	}
};

const focusNextItem = (host) => {
	const lastFocusItem = getFocusedItemEl(host);
	if (!lastFocusItem || isPanelList(lastFocusItem)) {
		focusFirstItem(host);
		return;
	}
	const lastFocusIndex = getFocusedItemIndex(host);
	const nextFocusItem = getItemEls(host)[lastFocusIndex + 1];
	if (nextFocusItem) {
		nextFocusItem.focus();
	}
};

const focusFirstItem = (host) => {
	const nextFocusItem = getItemEls(host)[0];
	if (nextFocusItem) {
		nextFocusItem.focus();
	}
};

const focusLastItem = (host) => {
	const nextFocusItem = getItemEls(host).slice(-1)[0];
	if (nextFocusItem) {
		nextFocusItem.focus();
	}
};
// end of focus

/** copy from ./utils/helps.js to isolate panel code*/
export const hasPath = (list, path) =>
	list.findIndex((p) => isEqual(p, path)) > -1;

export const removePath = (list, path) => list.filter((p) => !isEqual(p, path));

export const addPath = (list, path) => [...list, path];

export const togglePath = (list, path) =>
	hasPath(list, path) ? removePath(list, path) : addPath(list, path);
// end of helpers

const filterSelectedItems = (list, selectedItems) => {
	return selectedItems.map((item) => list.find((l) => item === l.id));
};

const renderCheckmark = (selected, select) => {
	if (select === 'none') {
		return;
	}
	if (selected) {
		return (
			<now-icon
				className="now-dropdown-panel-checkmark"
				icon="check-fill"
				aria-label={t('selected')}
			/>
		);
	}
	return <div className="now-dropdown-panel-checkmark" />;
};

const renderPresenceOrIcon = (
	icon,
	presence,
	selected,
	showPresencePattern
) => {
	if (presence) {
		return (
			<now-icon-presence
				className="now-m-inline-end--xxs"
				presence={presence}
				show-pattern={showPresencePattern}
			/>
		);
	} else if (icon) {
		return (
			<now-icon
				className="now-m-inline-end--xxs"
				icon={
					selected
						? icon.replace('outline', 'fill')
						: icon.replace('fill', 'outline')
				}
			/>
		);
	}
	return '';
};

const renderDropdownPanelItem = ({
	label,
	selected,
	select,
	onClick,
	key,
	icon,
	presence,
	showPresencePattern,
	count,
	disabled
}) => {
	return (
		<div
			tabindex="-1"
			id={key}
			class={{'now-dropdown-panel-item': true, 'is-disabled': disabled}}
			on-mouseover={(e) => {
				e.currentTarget.focus();
			}}
			on-mouseout={(e) => {
				const host = getShadowRootHost(e.currentTarget);
				unfocusItem(host);
			}}
			on-click={disabled ? undefined : onClick}
			role={select === 'none' ? 'menuitem' : 'option'}
			aria-selected={
				disabled || select === 'none' ? undefined : selected ? 'true' : 'false'
			}
			aria-disabled={disabled ? 'true' : undefined}>
			{renderCheckmark(selected, select)}
			{renderPresenceOrIcon(icon, presence, selected, showPresencePattern)}
			<span className="now-line-height-crop">
				<span className="now-align">
					<span
						className="now-dropdown-panel-label"
						on-mouseover={({target}) => {
							if (target.offsetWidth < target.scrollWidth) {
								target.title = label;
							}
						}}>
						{label}
					</span>
					{count ? `\u00A0(${count})` : undefined}
				</span>
			</span>
		</div>
	);
};

const view = (state, {dispatch}) => {
	const {componentId, fitStyle} = state;
	const {
		select,
		items,
		selectedItems,
		showPresencePattern,
		configAria
	} = state.properties;

	const onItemClick = (item) => {
		dispatch(() => {
			return {
				type: 'NOW_DROPDOWN_PANEL#ITEM_CLICKED',
				payload: {item}
			};
		});
		if (select === 'single' || select === 'multi') {
			dispatch(({properties}) => {
				const filteredItemIds = togglePath(selectedItems, item.id);
				return {
					type: 'NOW_DROPDOWN_PANEL#SELECTED_ITEMS_SET',
					payload: {
						value:
							properties.select === 'multi'
								? togglePath(properties.selectedItems, item.id)
								: [item.id],
						items:
							properties.select === 'multi'
								? filterSelectedItems(items, filteredItemIds)
								: [item]
					}
				};
			});
		}
		if (select !== 'multi') {
			dispatch(() => {
				return {
					type: 'NOW_DROPDOWN_PANEL#OPENED_SET',
					payload: {
						value: false,
						restoreFocus: true
					}
				};
			});
		}
	};

	const renderItem = (item) =>
		renderDropdownPanelItem({
			label: item.label,
			selected:
				selectedItems &&
				selectedItems.length &&
				selectedItems.includes(item.id),
			select,
			onClick: () => onItemClick(item),
			key: `${item.id}-${componentId}`,
			icon: item.icon,
			presence: item.presence,
			showPresencePattern,
			count: item.count,
			disabled: item.disabled
		});

	const renderSectionOrItem = (item, index, array) => {
		if (Array.isArray(item.children)) {
			const header = item.label ? (
				<div className="now-dropdown-panel-header">
					<div className="now-dropdown-panel-header-label" title={item.label}>
						{item.label}
					</div>
				</div>
			) : (
				undefined
			);
			const divider = canRenderDivider(index, array) ? (
				<hr className="now-dropdown-panel-divider" />
			) : (
				undefined
			);
			return (
				<Fragment>
					{header}
					{item.children.map(renderItem)}
					{divider}
				</Fragment>
			);
		}
		return renderItem(item);
	};

	return (
		<div
			id={`panel-${componentId}`}
			className="now-dropdown-panel"
			style={fitStyle}>
			<div
				className="now-dropdown-panel-list"
				role={select === 'none' ? 'menu' : 'listbox'}
				{...filterAriaAttributes(configAria, 'panel')}
				aria-multiselectable={select === 'multi' ? 'true' : undefined}>
				{items.map(renderSectionOrItem)}
			</div>
		</div>
	);
};

/**
 * The panel portion of a dropdown menu, for displaying a list of items.
 * This component should not be used directly in an application, but is used
 * by other higher level components like now-dropdown and now-split-button.
 *
 * @seismicElement now-dropdown-panel
 * @summary The panel portion of a dropdown menu, for displaying a list of items.
 *
 * */
createEnhancedElement('now-dropdown-panel', {
	properties: {
		/**
		 * An array of `DropdownItem` or `DropdownSection` items.
		 * @type {Array.<DropdownItem|DropdownSection>} // TODO: kind of a lie now, need to decide on deprecation of this component?
		 */
		items: {
			default: [],
			schema: itemsSchema // TODO: kind of a lie now, need to decide on deprecation of this component?
		},
		/**
		 * An array of IDs representing each selected item. If the multi-select mode
		 * is not enabled, the ID of a previously selected item is cleared when
		 * the user selects a new item, and a new ID is added. If the multi-select
		 * mode is enabled, IDs are added to or removed from the list.
		 * @type {Array.<(string|number)>}
		 */
		selectedItems: {
			default: [],
			manageable: true,
			schema: {type: 'array', items: {type: ['string', 'number']}}
		},
		/**
		 * Determines what happens when the user clicks on a dropdown panel item.
		 * Choose from the following:
		 * - "single" (Default): The panel is closed, the trigger label is updated,
		 *   and the item is marked as selected
		 * - "multi": The panel stays open, the trigger label is updated, and the
		 *   item's selected state is toggled on/off
		 * - "none": The panel is closed, the trigger label is not updated, and the
		 *   item is not marked as selected
		 * @type {("single"|"multi"|"none")}
		 */
		select: {
			default: 'single',
			schema: {type: 'string', enum: ['single', 'multi', 'none']}
		},
		/**
		 * Automatically updated when the user opens or closes the dropdown panel.
		 * Set `manage-opened` to override this behavior.
		 * @type {boolean}
		 */
		opened: {default: false, manageable: true, schema: {type: 'boolean'}},
		/**
		 * Reference to the element that defines the dropdown panel's relative
		 * position. The panel does not attach click/interaction handlers to this
		 * target. It is your responsibility to do so and to open/close the panel
		 * when the user clicks the trigger. The panel closes automatically when
		 * the user clicks outside or selects an item, depending on the select mode.
		 * @type {HTMLElement}
		 */
		targetRef: {},
		/**
		 * Pixel amount to shift the panel relative to its target.
		 *
		 * @see Fit behavior for valid values.
		 * @type {number}
		 * @private
		 */
		offset: {
			schema: {
				oneOf: [{type: 'number'}, {type: 'array', items: {type: 'number'}}]
			}
		},
		/**
		 * Element that the panel must fit within.
		 *
		 * @type {HTMLElement}
		 */
		container: {},
		/**
		 * Array of possible positions in which to place the panel,
		 * relative to its target.
		 *
		 * @see Fit behavior for valid values.
		 * @type {Array<string>}
		 */
		position: {
			default: [
				'bottom-start top-start',
				'top-start bottom-start',
				'bottom-end top-end',
				'top-end bottom-end'
			],
			schema: {
				type: 'array',
				items: {type: 'string'}
			}
		},
		/**
		 * Additional size constraints when sizing the panel,
		 * relative to its target.
		 *
		 * @see Fit behavior for valid values.
		 * @type {object}
		 */
		constrain: {
			default: {
				minHeight: 32,
				minWidth: 'target',
				maxWidth: 400,
				maxHeight: 300
			},
			schema: constrainSchema
		},
		/**
		 * When true, displays the dropdown item's presence status icon with a pattern to make
		 * it distinguishable by a means other than color.
		 * @type {boolean}
		 * @uib.description Visual pattern enabled on the presence indicator for accessibility
		 */
		showPresencePattern: {default: false, schema: {type: 'boolean'}},
		/**
		 * An object whose keys reference a specific HTML element within the
		 * now-dropdown. The value of each key is an object that contains an ARIA
		 * property (or properties) and corresponding value/s. These ARIA
		 * properties will be set on the specified element.
		 * The 'panel' key corresponds to the inner html `<div>` that acts as the
		 * 'listbox'. See https://www.w3.org/TR/wai-aria-1.1/#listbox for properties
		 * and accepted values.
		 * @type {{ 'aria-*': string }}
		 * @uib.label ARIA Properties
		 * @uib.description Sets ARIA properties on the inner div with role 'listbox'.
		 */
		configAria: {
			schema: {
				anyOf: [
					{
						type: 'object',
						properties: {
							panel: configAriaSchema
						}
					},
					configAriaSchema
				]
			}
		}
	},
	initialState: {
		fitStyle: {}
	},
	eventHandlers: [
		{
			events: ['click'],
			effect: ({action, dispatch, properties, host}) => {
				if (properties.opened) {
					const {event} = action.payload;
					const path = event.composedPath ? event.composedPath() : event.path;
					const hostInPath = path.includes(host);
					if (!hostInPath) {
						dispatch(() => {
							return {
								type: 'NOW_DROPDOWN_PANEL#OPENED_SET',
								payload: {
									value: false,
									restoreFocus: false
								}
							};
						});
					}
				}
			},
			target: document
		},
		{
			events: ['keydown'],
			effect({host, action, dispatch}) {
				const {key, shiftKey} = action.payload.event;

				if (['Home', 'ArrowUp', 'ArrowDown', 'End'].includes(key)) {
					action.preventDefault();

					if (key === 'Home') {
						focusFirstItem(host);
					}

					if (key === 'ArrowDown') {
						focusNextItem(host);
					}

					if (key === 'ArrowUp') {
						focusPreviousItem(host);
					}

					if (key === 'End') {
						focusLastItem(host);
					}
				} else if (['Escape', 'Tab'].includes(key)) {
					action.stopPropagation();
					const restoreFocus = key === 'Escape' || (key === 'Tab' && shiftKey);

					dispatch(() => {
						return {
							type: 'NOW_DROPDOWN_PANEL#OPENED_SET',
							payload: {value: false, restoreFocus}
						};
					});
				} else if ([' ', 'Enter'].includes(key)) {
					action.preventDefault();
					action.stopPropagation();

					const itemEl = getFocusedItemEl(host);

					if (itemEl && !itemEl.hasAttribute('aria-disabled')) {
						itemEl.click();
					}
				}
			}
		}
	],
	actionHandlers: {
		[COMPONENT_DOM_READY]: ({properties, host, state, dispatch}) => {
			const {targetRef, opened} = properties;
			if (targetRef && opened) {
				dispatch(() => {
					return {
						type: 'PRIVATE#REFIT',
						payload: {firstOpen: true}
					};
				});

				addOutsideScrollListener(host, state.componentId, () => {
					// Note: It is usually a VERY BAD IDEA to go reach into the component's
					// DOM and update things without going through the render cycle. In this
					// case we have to reach in and set the panel's visibility to hidden
					// immediately on scroll because dispatching will not trigger a render
					// fast enough and the panel will float there while the user scrolls.
					// It's ok. The component will re-render shortly after and set visibility
					// to hidden from state, so the DOM stays in sync with state.
					const panel = getPanelEl(host);
					if (panel) {
						panel.style.visibility = 'hidden';
					}
					dispatch('NOW_DROPDOWN#OPENED_SET', {value: false});
				});
			}
		},
		[COMPONENT_PROPERTY_CHANGED]: ({
			host,
			state,
			properties,
			action,
			dispatch
		}) => {
			const {
				payload: {name, value}
			} = action;

			if (
				(value && (name === 'opened' && properties.targetRef)) ||
				(name === 'targetRef' && properties.opened)
			) {
				addOutsideScrollListener(host, state.componentId, () => {
					// Note: It is usually a VERY BAD IDEA to go reach into the component's
					// DOM and update things without going through the render cycle. In this
					// case we have to reach in and set the panel's visibility to hidden
					// immediately on scroll because dispatching will not trigger a render
					// fast enough and the panel will float there while the user scrolls.
					// It's ok. The component will re-render shortly after and set visibility
					// to hidden from state, so the DOM stays in sync with state.
					const panel = getPanelEl(host);
					if (panel) {
						panel.style.visibility = 'hidden';
					}
					dispatch('NOW_DROPDOWN#OPENED_SET', {value: false});
				});
			} else if (!value && (name === 'opened' || name === 'targetRef')) {
				removeOutsideScrollListener(host);
			}

			if (
				name === 'items' ||
				name === 'container' ||
				name === 'constrain' ||
				name === 'targetRef' ||
				name === 'position' ||
				name === 'offset' ||
				name === 'opened'
			) {
				const firstOpen = name === 'opened';
				dispatch(() => {
					return {
						type: 'PRIVATE#REFIT',
						payload: {firstOpen}
					};
				});
			}
		},
		[COMPONENT_RENDERED]: ({state, host, updateState}) => {
			// TODO-SEISMIC-HACK: We want to focus the first item in the panel as the
			// panel becomes opened. The challenge here is that we don't know exactly
			// when the updateState that includes the fit changes has finished
			// rendering. The fit state update sets this `firstOpen` state that we
			// can immediately reset here so that we don't try to focus on every
			// render.
			if (state.firstOpen) {
				focusOnOpen(host);
				updateState({
					operation: 'set',
					path: 'firstOpen',
					value: false,
					shouldRender: false
				});
			}
		},
		'PRIVATE#REFIT': {
			stopPropagation: true,
			effect: ({action, state, properties, updateState, host}) => {
				const {
					targetRef: target,
					container,
					constrain,
					position,
					offset,
					opened
				} = properties;

				const content = getContentEl(host);
				const panelOpened = Boolean(opened && target);
				const fitStyle = panelOpened
					? {
							...getBestFitInfo({
								target,
								content,
								container,
								position,
								offset: Array.isArray(offset)
									? offset
									: [0, 0, offset, offset, offset, offset],
								constrain,
								canScrollVertical: true
							}).style,
							visibility: 'visible'
					  }
					: {};
				if (!isEqual(fitStyle, state.fitStyle)) {
					updateState({
						fitStyle,
						firstOpen: action.payload.firstOpen
					});
				}
			}
		}
	},
	view,
	styles,
	dispatches: {
		/**
		 * Dispatched when the user opens or closes the panel. The payload contains
		 * a boolean of `true` when the dropdown is open or `false` when it is
		 * closed. Set the `manage-opened` property to override
		 * the default behavior and handle this action manually.
		 * @type {{value: boolean}}
		 */
		'NOW_DROPDOWN_PANEL#OPENED_SET': {},
		/**
		 * Dispatched when the user changes the selection of items in the dropdown panel.
		 * The payload key `value` contains an array containing either a single ID (`select="single"`)
		 * or multiple IDs (`select="multi"`) of the selected item(s).
		 * The payload key `items` contains an array of objects of the item or items selected from the dropdown list.
		 * Set the `manage-selected-items` property to override the default behavior and
		 * handle this action manually.
		 * @type {{value: Array.<(string|number)>, items: Array.<Object>}}
		 */
		'NOW_DROPDOWN_PANEL#SELECTED_ITEMS_SET': {},
		/**
		 * Dispatched when the user selects a dropdown item. The payload contains
		 * a reference to the clicked item. Handle this action for dropdown menus
		 * (where `select="none"`).
		 * @type {{item: DropdownItem}}
		 */
		'NOW_DROPDOWN_PANEL#ITEM_CLICKED': {}
	}
});
