import {createEnhancedElement} from '@servicenow/library-enhanced-element';
import {
	extensionSplit,
	filterAriaAttributes,
	makeRedispatchEffect
} from '@servicenow/library-enhanced-utils';
import styles from './_now-dropdown.scss';
import {getFocusedEl, getTriggerEl} from './utils/elements.js';
import {setActiveItem, getDropdownListConstrain} from './utils/helpers.js';
import focusBehavior from '@servicenow/behavior-focus';
import '@servicenow/now-icon';
import '@servicenow/now-tooltip';
import '@servicenow/now-popover';
import './now-dropdown-list';
import {t} from '@servicenow/library-translate';
import {itemsSchema, itemSchema} from './schemas/dropdown-schema';
import {constrainSchema} from '@servicenow/behavior-fit';
import {pick} from 'lodash';
import {configAriaSchema} from '@servicenow/library-schemas';

const getSelected = (items, selectedItemId) => {
	const hasNestedChildren = items.some((item) =>
		item.hasOwnProperty('children')
	);
	if (hasNestedChildren) {
		return items
			.reduce(
				(acc, item) =>
					Array.isArray(item.children)
						? [...acc, ...item.children]
						: [...acc, item],
				[]
			)
			.find((item) => item.id === selectedItemId);
	} else {
		return items.find((item) => selectedItemId === item.id);
	}
};

const getTriggerLabel = (placeholder, selected, selectedItems) => {
	if (Array.isArray(selectedItems) && selectedItems.length > 1) {
		return t('{0} items selected', selectedItems.length);
	} else if (selected) {
		return selected.label;
	}
	return placeholder;
};

const getTriggerIcon = (icon, selected) => {
	if (selected && selected.icon) {
		return selected.icon.replace('outline', 'fill');
	}
	if (selected && selected.presence) {
		return 'presence-fill';
	}
	return icon;
};

const renderIcon = (presence, showPresencePattern, size, hasLabel, icon) => {
	if (presence) {
		return (
			<now-icon-presence
				presence={presence}
				show-pattern={showPresencePattern}
				size={size}
				class={{
					'now-m-inline-end--xxs': hasLabel && size !== 'lg',
					'now-m-inline-end--xs': hasLabel && size === 'lg'
				}}
				aria-hidden="true"
			/>
		);
	} else if (icon) {
		return (
			<now-icon
				icon={icon}
				size={size}
				class={{
					'now-m-inline-end--xxs': hasLabel && size !== 'lg',
					'now-m-inline-end--xs': hasLabel && size === 'lg'
				}}
				aria-hidden="true"
			/>
		);
	}
};

const renderTooltip = (tooltipContent, componentId, fitTarget) => {
	if (tooltipContent) {
		return (
			<now-tooltip
				id={`tooltip-${componentId}`}
				target-ref={fitTarget}
				content={tooltipContent}
			/>
		);
	}
};

const renderLabel = (hasLabel, label, icon) => {
	if (hasLabel) {
		return (
			<span
				class={{
					'now-dropdown-label': true,
					'now-m-inline-start--xxs': icon
				}}
				title={label}
				aria-hidden="true">
				{label}
			</span>
		);
	}
};

const renderCaret = (hideCaret, hasLabel, size) => {
	if (!hideCaret) {
		return (
			<now-icon
				class={{
					'now-m-inline-start--xxs': hasLabel && size !== 'lg',
					'now-m-inline-start--xs': hasLabel && size === 'lg'
				}}
				size={size}
				icon="caret-down-fill"
				aria-hidden="true"
			/>
		);
	}
};

const view = (state, {updateState}) => {
	const {componentId, fitTarget, activeItem} = state;
	const {
		opened,
		select,
		items,
		selectedItems,
		placeholder,
		icon,
		disabled,
		variant,
		size,
		bare,
		hideCaret,
		tooltipContent,
		panelFitProps = {},
		hideLabel,
		showPadding,
		showPresencePattern,
		configAria,
		search
	} = state.properties;

	const setFitTarget = (ref) => setTimeout(() => updateState({fitTarget: ref}));
	const constrain = {
		minHeight: 32,
		minWidth: 'target',
		maxWidth: 400,
		maxHeight: 300,
		...panelFitProps.constrain
	};
	let selected;
	if (Array.isArray(selectedItems) && selectedItems.length === 1) {
		selected = getSelected(items, selectedItems[0]);
	}
	const label = getTriggerLabel(placeholder, selected, selectedItems);
	const hasLabel = label && !hideLabel;
	const triggerIcon = getTriggerIcon(icon, selected);
	const presence =
		selected && selected.presence ? selected.presence : undefined;
	const count = selected && selected.count ? selected.count : undefined;

	return (
		<now-popover
			hide-tail
			constrain={pick(constrain, ['width', 'minWidth', 'maxWidth'])}
			positions={panelFitProps.position}
			offset={panelFitProps.offset}
			manage-opened
			opened={opened}
			interaction-type="listbox">
			<button
				slot="trigger"
				type="button"
				id={`trigger-${componentId}`}
				class={{
					'now-dropdown': true,
					...(!disabled ? extensionSplit(variant) : {}),
					['-' + size]: size,
					'is-bare':
						bare &&
						(variant === 'secondary' ||
							variant === 'secondary-selected' ||
							variant === 'tertiary' ||
							variant === 'inherit'),
					'is-disabled': disabled,
					'has-label': !hideLabel && label,
					'has-caret': !hideCaret,
					'is-active': opened,
					'show-padding': showPadding
				}}
				hook-insert={({elm}) => setFitTarget(elm)}
				hook-destroy={() => setFitTarget(null)}
				aria-label={label}
				{...filterAriaAttributes(configAria, 'trigger')}
				aria-haspopup="listbox"
				aria-expanded={opened ? 'true' : 'false'}
				aria-disabled={disabled ? 'true' : 'false'}
				aria-describedby={tooltipContent ? `tooltip-${componentId}` : undefined}
				tabindex={disabled && !tooltipContent ? '-1' : '0'}
				on-mousedown={disabled ? (e) => e.preventDefault() : undefined}>
				<slot name="trigger-content">
					{renderIcon(
						presence,
						showPresencePattern,
						size,
						hasLabel,
						triggerIcon
					)}
					<span className="now-line-height-crop">
						<span className="now-align">
							{renderLabel(hasLabel, label, icon)}
							{count ? `\u00A0(${count})` : undefined}
						</span>
					</span>
					{renderCaret(hideCaret, hasLabel, size)}
				</slot>
				{renderTooltip(tooltipContent, componentId, fitTarget)}
			</button>
			<now-dropdown-list
				slot="content"
				style={getDropdownListConstrain(constrain)}
				items={items}
				selected-items={selectedItems}
				select={select}
				search={search}
				show-presence-pattern={showPresencePattern}
				active-item={activeItem}
				manage-selected-items
				manage-active-item
				config-aria={
					placeholder || tooltipContent
						? {
								panel: {'aria-label': placeholder || tooltipContent},
								...configAria
						  }
						: configAria
				}
			/>
		</now-popover>
	);
};

/**
 * A dropdown is a partially hidden panel element that is used to display a list of menu options.
 * A user selects the dropdown to expose the list of options hidden in the panel. This gives users
 * the flexibility to make one or more selections from the list of options that appear in the panel.
 *
 * ```jsx
 * const dropdownItems = [{id: 'item1', label: 'Option 1'}, {id: 'item2', label: 'Option 2'}];
 *
 * <now-dropdown placeholder="Select" items={dropdownItems} />
 *
 * <now-dropdown
 *    items={dropdownItems}
 *    icon="tag-fill"
 *    tooltip-content="Select an item"
 *    config-aria={{
 * 		trigger: {'aria-label': 'Select an item'},
 * 		panel: {'aria-label': 'Assigned to'}
 * 	  }}
 *    bare />
 *
 * <now-dropdown items={dropdownItems} bare select="none">
 *   <now-avatar slot="trigger-content" user-name="Manuel Bernada" />
 * </now-dropdown>
 *
 * const dropdownItemsWithDividers = [{label: 'Section 1', children: dropdownItems}];
 * <now-dropdown items={dropdownItemsWithDividers} />
 * ```
 *
 * > NOTE: `now-dropdown-panel` is deprecated as of version `19.0.12`. We recommend switching to `now-dropdown-custom-target` instead.
 *
 *
 * @seismicElement now-dropdown
 * @summary A dropdown menu that allows user to select one or more values.
 *
 * @uib.label Dropdown
 * @uib.icon dropdown
 * @uib.description Lets users select an option from a list in a panel
 * @uib.category primitives
 * @uib.associatedTypes global.core
 * @uib.properties variant, select, size, placeholder, hideLabel, icon, bare, showPadding, hideCaret, disabled, showPresencePattern, items, opened, panelFitProps, selectedItems, configAria, tooltipContent
 * @uib.actions NOW_DROPDOWN#OPENED_SET, NOW_DROPDOWN#SELECTED_ITEMS_SET, NOW_DROPDOWN#ITEM_CLICKED
 *
 * */
createEnhancedElement('now-dropdown', {
	properties: {
		/**
		 * An array of `DropdownItem` or `DropdownSection` items.
		 * @type {(Array.<DropdownItem> | Array.<DropdownSection>)}
		 * @uib.defaultValue [{"id":"l1","label":"List item 1"}, {"id":"l2","label":"List item 2"}, {"id":"l3","label":"List item 3"}, {"id":"l4","label":"List item 4"}, {"id":"l5","label":"List item 5"}, {"id":"l6","label":"List item 6"}]
		 * @uib.description An array of dropdown items or sections
		 * @uib.typeMetadata.emptyStateBehavior EMPTY_ARRAY
		 */
		items: {
			required: true,
			schema: itemsSchema
		},
		/**
		 * An array of ids representing selected dropdown panel items. If multi-select mode is
		 * not enabled, each time the user selects a list item, the previously selected
		 * id will be replaced with the new id. If multi-select mode is
		 * enabled, ids will be added to or removed from the list.
		 * Use `manage-selected-items` to override the default behavior and handle the
		 * `NOW_DROPDOWN#SELECTED_ITEMS_SET` action manually.
		 * @type {Array.<(string|number)>}
		 * @uib.description An array of ids representing selected dropdown panel items. If multi-select mode is not enabled, each time the user selects a list item, the previously selected id will be replaced with the new id. If multi-select mode is enabled, ids will be added to or removed from the list.
		 */
		selectedItems: {
			default: [],
			manageable: true,
			schema: {type: 'array', items: {type: ['string', 'number']}}
		},
		/**
		 * Determines what happens when the user selects a dropdown item.
		 * Choose from the following:
		 * - "single" (default) - the panel is closed, the trigger label is updated with
		 *   the label of the selected list item, and the list item is marked as selected.
		 * - "multi" - the panel stays open, the trigger label is updated with the selected
		 *    list item count, and the list item's selected state is toggled on/off.
		 * - "none" - the panel is closed, the trigger label is not updated, and the
		 *   list item is not marked as selected.
		 * When using the "single" or "multi" option, handle the `NOW_DROPDOWN#SELECTED_ITEMS_SET`
		 * action manually. When using the "none" option, handle the `NOW_DROPDOWN#ITEM_CLICKED` action manually.
		 * @type {("single"|"multi"|"none")}
		 * @uib.label Selection
		 * @uib.description Determines what happens when the user selects a dropdown item.
		 */
		select: {
			default: 'single',
			schema: {type: 'string', enum: ['single', 'multi', 'none']}
		},
		/**
		 * Automatically updates when the user opens or closes the dropdown.
		 * Use `manage-opened` to override the default behavior and handle the
		 * `NOW_DROPDOWN#OPENED_SET` action manually.
		 * @type {boolean}
		 * @uib.description Automatically updates when the user opens or closes the dropdown.
		 */
		opened: {default: false, manageable: true, schema: {type: 'boolean'}},
		/**
		 * The fallback label displays when no items are selected. If you do not
		 * specify a placeholder value, especially in the case of iconic
		 * dropdowns, provide `tooltipContent` and/or an `aria-label`
		 * via the `configAria` properties.
		 * @type {string}
		 * @uib.label Placeholder text
		 * @uib.description The fallback label that displays when no items are selected. If you do not specify a placeholder value, especially in the case of iconic dropdowns, provide tooltipContent and/or an aria-label via the configAria properties.
		 * @uib.defaultValue Dropdown
		 * @uib.translatable true
		 */
		placeholder: {schema: {type: 'string'}},
		/**
		 * Enable to show the dropdown as visually disabled and to stop the dropdown
		 * from opening when the user interacts with it.
		 * @type {boolean}
		 * @uib.label Disabled
		 * @uib.description Disables user click interactions
		 */
		disabled: {default: false, schema: {type: 'boolean'}},
		/**
		 * Placeholder icon to include in the dropdown trigger. This icon will be replaced by
		 * the icon of a selected dropdown item if the item has an icon configured.
		 * @type {string}
		 * @uib.label Placeholder icon
		 * @uib.description Placeholder icon name to include in the dropdown trigger. This icon is replaced by the icon of a selected dropdown item if the item has an icon configured
		 * @uib.fieldType icon
		 */
		icon: {schema: {type: 'string'}},
		/**
		 * Sets the button styles of the dropdown trigger.
		 * @type {("primary"|"secondary"|"secondary-selected"|"tertiary"|"tertiary-selected"|"inherit")}
		 * @uib.label Variant
		 * @uib.description Sets the dropdown styles.
		 * @uib.typeMetadata.choices [{"label": "Primary","value": "primary"},{"label": "Secondary","value": "secondary"},{"label": "Secondary (selection)","value": "secondary-selected"},{"label": "Tertiary","value": "tertiary"},{"label": "Tertiary (selection)","value": "tertiary-selected"},{"label": "Inherit","value": "inherit"}]
		 */
		variant: {
			default: 'secondary',
			schema: {
				type: 'string',
				enum: [
					'primary',
					'secondary',
					'secondary-selected',
					'tertiary',
					'tertiary-selected',
					'inherit'
				]
			}
		},
		/**
		 * Sets the button size of the dropdown trigger.
		 * @type {("sm"|"md"|"lg")}
		 * @uib.label Size
		 * @uib.description Sets the button size of the dropdown trigger.
		 */
		size: {default: 'md', schema: {type: 'string', enum: ['sm', 'md', 'lg']}},
		/**
		 * Makes the dropdown trigger border and background transparent, but retains
		 * its `variant` style setting when using an iconic button. If button is text or
		 * text with an icon then only works when `variant` is `secondary`,
		 * `tertiary`, or `inherit`.
		 * @type {boolean}
		 * @uib.label Hide Container
		 * @uib.description Hides the dropdown container and retains the style when using an iconic button. If the button is text or text with an icon, then this setting only applies when variant is secondary, tertiary, or inherit.
		 */
		bare: {default: false, schema: {type: 'boolean'}},
		/**
		 * Hides the triangular caret icon that appears in the dropdown trigger by
		 * default.
		 * @type {boolean}
		 * @uib.description Hides the caret icon.
		 */
		hideCaret: {default: false, schema: {type: 'boolean'}},
		/**
		 * Text shown inside the tooltip
		 * @type {string}
		 * @uib.label Tooltip text
		 * @uib.description Text content shown inside the tooltip.
		 * @uib.translatable true
		 */
		tooltipContent: {schema: {type: 'string'}},
		/**
		 * Properties to configure the dropdown panel.
		 *
		 * Valid keys are:
		 *	- position
		 *	- constrain
		 *  - offset
		 *
		 * Each of these keys are used to control how the panel is positioned and
		 * scaled relative to the dropdown target.
		 *
		 * @type {{ position: Array<{target: string, content: string}>, constrain: object, offset: number }}
		 * @uib.label Panel Position
		 * @uib.description Positions and scales the panel relative to the dropdown trigger using 'position', 'container', or 'constrain'
		 */
		panelFitProps: {
			default: {},
			schema: {
				type: 'object',
				properties: {
					position: {type: 'array', items: {type: 'string'}},
					offset: {
						oneOf: [{type: 'number'}, {type: 'array', items: {type: 'number'}}]
					},
					constrain: constrainSchema
				}
			}
		},
		/**
		 * Set this flag to suppress a visual indication of
		 * the selected item(s). By default, the dropdown displays
		 * the selected item or a count of selected items in place of the
		 * `placeholder` text in the label of the dropdown trigger.
		 * @type {boolean}
		 * @uib.label Hide selected item
		 * @uib.description Suppresses a visual indication of the selected item(s). By default, the dropdown displays the selected item or a count of selected items in place of the placeholder text in the label of the dropdown trigger.
		 */
		hideLabel: {default: false, schema: {type: 'boolean'}},
		/**
		 * Set this flag to always include side padding on bare buttons.
		 * By default, bare buttons with labels do not have side padding.
		 * @type {boolean}
		 * @uib.description Includes side padding on bare buttons. By default, bare buttons with labels do not have side padding.
		 */
		showPadding: {default: false, schema: {type: 'boolean'}},
		/**
		 * When true, displays the dropdown item's presence status icon with a pattern to make
		 * it distinguishable by a means other than color.
		 * @type {boolean}
		 * @uib.description Enables a visual pattern on the presence indicator for accessibility
		 */
		showPresencePattern: {default: false, schema: {type: 'boolean'}},
		/**
		 * An object whose keys reference a specific HTML element within the
		 * now-dropdown. The value of each key is an object that contains an ARIA
		 * property (or properties) and corresponding value/s. These ARIA
		 * properties will be set on the specified element.
		 * The 'trigger' key corresponds to the inner html `<button>` that acts as
		 * the trigger in now-dropdown. See https://www.w3.org/TR/wai-aria-1.1/#button
		 * for properties and accepted values.
		 * The 'panel' key corresponds to the inner html `<div>` that acts as the
		 * 'listbox'. See https://www.w3.org/TR/wai-aria-1.1/#listbox for properties
		 * and accepted values.
		 * @type {{ 'aria-*': string }}
		 * @uib.label ARIA Properties
		 * @uib.description Sets ARIA properties on the inner html 'button' and 'listbox'.
		 */
		configAria: {
			schema: {
				oneOf: [
					{
						type: 'object',
						properties: {
							trigger: configAriaSchema,
							panel: configAriaSchema
						},
						additionalProperties: false
					},
					configAriaSchema
				]
			}
		},
		/**
		 * Controls how to filter the items as the user types into the integrated search input.
		 * - "none": (default) dropdown list does not display an internal search field.
		 * - "managed": internal input field is rendered, but no automatic filtering will happen.
		 * Use this if you want to manually filter or search for items as the user types outside of the list.
		 * - "initial": internal input field is rendered, and items whose label
		 * or sublabel begins with the input value will be displayed. This is case-insensitive.
		 * - "contains": internal input field is rendered, and items whose label
		 * or sublabel contains the input value will be displayed. This is case-insensitive.
		 * @type {("none"|"managed"|"initial"|"contains")}
		 * @uib.label Search Behavior
		 * @uib.description Controls how to filter the items as the user types into the integrated search input
		 */
		search: {
			default: 'none',
			schema: {
				type: 'string',
				enum: ['none', 'managed', 'initial', 'contains']
			}
		}
	},
	dispatches: {
		/**
		 * Dispatched when the dropdown menu is opened or closed. Payload contains
		 * a boolean of `true` when the dropdown is opened or `false` when the dropdown is closed.
		 * Set the `manage-opened` property to override the default behavior and
		 * handle this action manually.
		 * @type {{value: boolean}}
		 * @uib.description Dispatched when the dropdown menu is opened or closed
		 */
		'NOW_DROPDOWN#OPENED_SET': {
			schema: {
				type: 'object',
				properties: {
					value: {type: 'boolean'}
				}
			}
		},
		/**
		 * Dispatched when the user changes the selection of items in the dropdown panel.
		 * Payload contains an array containing either a single ID (for `select="single"`) or
		 * multiple IDs (for `select="multi"`) of the selected item(s).
		 * Set the `manage-selected-items` property to override the default behavior and
		 * handle this action manually.
		 * @type {{value: Array.<(string|number)>}}
		 * @uib.description Dispatched when the user changes the selection of items in the dropdown panel for Selection mode `single` and `multi`
		 */
		'NOW_DROPDOWN#SELECTED_ITEMS_SET': {
			schema: {
				type: 'object',
				properties: {
					value: {
						type: 'array',
						items: {
							type: ['string', 'number']
						}
					}
				}
			}
		},
		/**
		 * Dispatched when the user selects a dropdown item. Payload contains
		 * a reference to the clicked item. Handle this action for dropdown menus (where `select="none"`).
		 * @type {{item: DropdownItem}}
		 * @uib.description Dispatched when the user selects a dropdown item for Selection mode `none`
		 */
		'NOW_DROPDOWN#ITEM_CLICKED': {
			schema: {
				type: 'object',
				properties: {item: itemSchema}
			}
		}
	},
	slots: {
		/**
		 * Content placed in this named slot appears inside of the trigger button of the dropdown,
		 * taking the place of the `icon`, `placeholder` and caret icon that normally appear there.
		 * Click behavior and other functionality is maintained for the custom content automatically.
		 */
		'trigger-content': {}
	},
	eventHandlers: [
		{
			events: ['keydown'],
			effect({host, action, properties, dispatch, updateState}) {
				const {key} = action.payload.event;
				const {items, selectedItems} = properties;

				if (['ArrowDown', 'ArrowUp'].includes(key)) {
					const focused = getFocusedEl(host);
					const trigger = getTriggerEl(host);

					if (!properties.disabled && focused === trigger) {
						action.preventDefault();
						setActiveItem(items, selectedItems, updateState);
						dispatch(() => {
							return {
								type: 'NOW_DROPDOWN#OPENED_SET',
								payload: {value: true}
							};
						});
					}
				}
			}
		}
	],
	actionHandlers: {
		'NOW_POPOVER#OPENED_SET': ({action, dispatch, properties}) => {
			const {value} = action.payload;
			const {disabled} = properties;
			if (disabled) return;
			dispatch(() => {
				return {
					type: 'NOW_DROPDOWN#OPENED_SET',
					payload: {value}
				};
			});
		},
		'NOW_POPOVER#CONTENT_VISIBLE': ({host, properties, updateState}) => {
			const dropdownList = host.shadowRoot.querySelector('now-dropdown-list');
			dropdownList.focus();

			const {items, selectedItems} = properties;
			setActiveItem(items, selectedItems, updateState);
		},
		'NOW_POPOVER#CONTENT_HIDDEN': ({updateState}) => {
			updateState(() => {
				return {
					path: 'activeItem',
					value: null,
					operation: 'set',
					shouldRender: false
				};
			});
		},
		'NOW_DROPDOWN_LIST#SELECTED_ITEMS_SET': makeRedispatchEffect(
			'NOW_DROPDOWN#SELECTED_ITEMS_SET'
		),
		'NOW_DROPDOWN_LIST#ITEM_CLICKED': ({dispatch, action}) => {
			dispatch(({properties}) => {
				const {select} = properties;
				return {
					type: 'NOW_DROPDOWN#OPENED_SET',
					payload: {value: false},
					shouldDispatch: select !== 'multi'
				};
			});
			dispatch(() => ({
				type: 'NOW_DROPDOWN#ITEM_CLICKED',
				payload: {item: action.payload.item}
			}));
		},
		'NOW_DROPDOWN_LIST#ACTIVE_ITEM_SET': ({action, updateState}) => {
			const activeItem = action.payload.value;
			updateState({activeItem});
		}
	},
	behaviors: [{behavior: focusBehavior}],
	view,
	styles
});
